// Solutions to the Advent Of Code 2024.
// Copyright (C) 2025  Stefan Müller
//
// This program is free software: you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free Software
// Foundation, either version 3 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along with
// this program.  If not, see <http://www.gnu.org/licenses/>.

#include <aoc/ResonantCollinearity.hpp>

#include <iostream>

const std::string ResonantCollinearity::getPuzzleName() const
{
    return "Day 8: Resonant Collinearity";
}

const std::string ResonantCollinearity::getInputFileName() const
{
    return "resonant_collinearity.txt";
}

void ResonantCollinearity::finish()
{
    for (int j = 0; j < lines.size(); j++)
    {
        for (int i = 0; i < lines[j].size(); i++)
        {
            if (lines[j][i] != getEmptyChar())
            {
                addNewAntenna({ i, j });
            }
        }
    }

    part1 = antinodes_.size();
}

constexpr char ResonantCollinearity::getEmptyChar()
{
    return '.';
}

void ResonantCollinearity::addNewAntenna(Point2&& newPosition)
{
    const auto [it, success] = antennas_.insert({ getPosition(newPosition), { newPosition } });
    if (!success)
    {
        for (const auto& oldPosition : it->second)
        {
            addNewAntinode(newPosition * 2 - oldPosition);
            addNewAntinode(oldPosition * 2 - newPosition);
        }

        it->second.push_back(newPosition);
    }
}

void ResonantCollinearity::addNewAntinode(Point2&& newPosition)
{
    if (isInBounds(newPosition))
    {
        antinodes_.insert(newPosition);
    }
}
